@@ -5,7 +5,7 @@ import os
 from .parse import parse_json, parse_essence
 from .mus import CascadeMUSFinder, checkWhichLitsAMUSProves
 from .musforqes import ForqesMUSFinder
-from .utils import flatten, intsqrt, lowsqrt
+from .utils import flatten, in_flattened, intsqrt, lowsqrt
 from .base import EqVal, NeqVal
 
 from sortedcontainers import SortedSet
@@ -22,8 +22,8 @@ class ExplainError(Exception):
 
 
 class Explainer(object):
-    def __init__(self, mus_finder=None, merge=1, skip=0, debug=False):
-        self.steps_explained = 0
+    def __init__(self, mus_finder=None, merge=1, skip=0, debug=False, steps_explained=0):
+        self.steps_explained = steps_explained
         self.mus_finder_name = mus_finder
         self.merge = merge
         self.skip = skip
@@ -38,6 +38,7 @@ class Explainer(object):
         self.puzzle = None
         self.solver = None
         self.solution = None
+        self.explained = []
 
     def init_from_json(self, puzzle_json):
         self.puzzle, self.solver = parse_json(puzzle_json)
@@ -166,7 +167,7 @@ class Explainer(object):
     def get_choices(self):
         mus_dict = self.mus_finder.smallestMUS(self.unexplained)
         choices_explanations, _ = self._choices_list(mus_dict)
-        return choices_explanations
+        return {"name": self.name, "params": self.params, "steps": [choices_explanations]}
 
     def _choices_list(self, mus_dict):
         smallest = mus_dict.minimum()
@@ -199,6 +200,7 @@ class Explainer(object):
             # Tell we solver we know this
             self.solver.addLit(p)
             # Remove from the things we have to calculate
+            self.explained.append(p)
             self.unexplained.remove(p)
 
     def _get_step_dict(self, proven_lits, mus):
@@ -308,9 +310,7 @@ class Explainer(object):
 
                 for i, clause in enumerate(involved):
 
-                    if (poslit in flatten(clause)) or (
-                        neglit in flatten(clause)
-                    ):
+                    if in_flattened(clause, poslit) or in_flattened(clause, neglit):
 
                         explanations.append(str(i))
                         # We want this to be "the" explanation that makes d
@@ -360,6 +360,8 @@ class Explainer(object):
         best_mus_stat = (math.inf, math.inf, math.inf)
         proven_dict = {}
 
+        checked = set([])
+
         for b in candidates:
             proven_dict[b] = {}
             for mus in mus_dict.get(b):
@@ -368,6 +370,11 @@ class Explainer(object):
                     SortedSet(), *(SortedSet(m.lits()) for m in mus)
                 )
 
+                # TODO: Filter out duplicated MUSes
+                #if mus in checked:
+                #    continue
+                #checked.add(mus)
+
                 unexplained_in_mus = SortedSet(
                     p
                     for p in self.unexplained
@@ -376,11 +383,14 @@ class Explainer(object):
 
                 # Explictly add 'b', for the case where the MUS is size 0 in
                 # particular
-                proven_lits = SortedSet(
-                    checkWhichLitsAMUSProves(
-                        self.solver, unexplained_in_mus, mus
-                    )
-                ).union(SortedSet([b]))
+                if (len(mus), len(unexplained_in_mus)) < best_mus_stat:
+                    proven_lits = SortedSet(
+                        checkWhichLitsAMUSProves(
+                            self.solver, unexplained_in_mus, mus
+                        )
+                    ).union(SortedSet([b]))
+                else:
+                    proven_lits = SortedSet([b])
 
                 proven_dict[b][mus] = proven_lits
 
