@@ -3,6 +3,7 @@ import math
 import logging
 import sys
 import math
+import multiprocessing
 
 from sortedcontainers import *
 
@@ -14,14 +15,14 @@ from .musdict import MusDict
 # This calculates Minimum Unsatisfiable Sets
 # It uses internals from solver, but is put in another file just for "neatness"
 
-def tinyMUS(solver, assume, distance):
+def tinyMUS(solver, assume, distance, badlimit):
     smtassume = [solver._varlit2smtmap[l] for l in assume]
     if distance == 1:
         cons = flatten([solver._varlit2con[l] for l in assume])
     elif distance == 2:
         cons = flatten([solver._varlit2con2[l] for l in assume])
     else:
-        sys.exit(1)
+        cons = list(solver._conlits)
 
     core = solver.basicCore(smtassume + cons)
     if core is None:
@@ -38,9 +39,11 @@ def tinyMUS(solver, assume, distance):
                 core = newcore
             else:
                 badcount += 1
-                if badcount > 5:
+                if badcount > badlimit:
+                    logging.info("ZZFail %s %s %s", lit, len(core), badcount)
                     return None
 
+    logging.info("ZZPass %s %s %s", lit, len(core), badcount)
     return [solver._conmap[x] for x in core if x in solver._conmap]
 
 
@@ -123,8 +126,8 @@ def MUS(
                 break
 
         logging.debug(
-            "tryManyChop: %s %s %s %s %s",
-            squash,
+            "tryManyChop: %s %s %s %s %s %s",
+            assume, squash,
             step,
             loopsize,
             len(core),
@@ -133,7 +136,7 @@ def MUS(
 
         if loopsize <= 10:
             done = False
-            for tries in range(loopsize * 2):
+            for tries in range(loopsize):
                 r.shuffle(core)
                 newcore = solver.basicCore(smtassume + core[:-step])
                 if newcore is not None:
@@ -322,14 +325,14 @@ def MUS(
                 core = newcore
                 lens.append((lit, len(core)))
             else:
-                logging.debug("Failed to remove: %s", lit)
+                logging.debug("Failed to remove: %s (%d of %d)", lit, badcount, minsize)
                 badcount += 1
 
                 if badcount == minsize:
                     cutcore = core[:minsize]
                     # Check if the core is already minimal first
                     if cutcore != core and (
-                        solver._solver.solve(smtassume + to_test, getsol=False)
+                        solver._solver.solve(smtassume + cutcore, getsol=False)
                         != False
                     ):
                         logging.debug(
@@ -341,20 +344,21 @@ def MUS(
                         )
                         return None
                     else:
-                        logging.debug(
-                            "Core found: %s %s %s %s",
+                        logging.info(
+                            "Core found by badcount: %s %s %s %s %s",
                             assume,
                             minsize,
                             badcount,
                             stepcount,
+                            len(cutcore)
                         )
                         return [
                             solver._conmap[x]
-                            for x in to_test
+                            for x in cutcore
                             if x in solver._conmap
                         ]
 
-    logging.debug(
+    logging.info(
         "Core for %s : %s to %s, with %s steps, %s bad (minsize %s)",
         assume,
         lens,
@@ -367,11 +371,11 @@ def MUS(
 
 
 def _parfunc_dotinymus(args):
-    (p, distance) = args
-    return (p, tinyMUS(getChildSolver(), [p.neg()], distance))
+    (p, distance, badlimit) = args
+    return (p, tinyMUS(getChildSolver(), [p.neg()], distance, badlimit))
 
 
-def getTinyMUSes(solver, puzlits, musdict, *, distance, repeats):
+def getTinyMUSes(solver, puzlits, musdict, *, distance, repeats, badlimit):
     setChildSolver(solver)
     logging.info(
         "Getting tiny MUSes, distance %s, for %s puzlits, %s repeats",
@@ -382,7 +386,7 @@ def getTinyMUSes(solver, puzlits, musdict, *, distance, repeats):
     with getPool(CONFIG["cores"]) as pool:
         res = pool.map(
             _parfunc_dotinymus,
-            [(p, distance) for r in range(repeats) for p in puzlits],
+            [(p, distance, badlimit) for r in range(repeats) for p in puzlits],
         )
         for (p, mus) in res:
             musdict.update(p, mus)
@@ -445,10 +449,22 @@ def checkWhichLitsAMUSProves(solver, puzlits, mus):
         return []
 
 
+MUSSizeFound = None
+MUSSizeRequired = None
+
+MAX_MUS = 999999999
+
 def _findSmallestMUS_func(tup):
     (p, randstr, minsize, config) = tup
+
+    logging.info("YY %s %s %s %s", MUSSizeFound.value, MUSSizeRequired.value, minsize, p)
+
+    if CONFIG["earlyExit"] and MUSSizeFound.value <= MUSSizeRequired.value:
+        logging.info("Early Exit!")
+        return (p, None)
+
     # logging.info("Random str: '%s'", randstr)
-    return (
+    (ret, mus) = (
         p,
         MUS(
             randomFromSeed(randstr),
@@ -458,17 +474,36 @@ def _findSmallestMUS_func(tup):
             config=config,
         ),
     )
+    if mus is not None:
+        if len(mus) < MUSSizeFound.value:
+            logging.info("Found new best MUS size: %s -> %s", MUSSizeFound.value, len(mus))
+            MUSSizeFound.value = len(mus)
+    return (ret, mus)
+
 
 
 def cascadeMUS(solver, puzlits, repeats, musdict, config):
     # We need this to be accessible by the pool
     setChildSolver(solver)
+    global MUSSizeFound, MUSSizeRequired
+    if musdict.minimum() < math.inf:
+        MUSSizeFound = multiprocessing.Value('l', musdict.minimum())
+    else:
+        MUSSizeFound = multiprocessing.Value('l', MAX_MUS)
+ 
+    MUSSizeRequired = multiprocessing.Value('l', 111)
 
     # Have to duplicate code, to swap loops around
     if CONFIG["resetSolverMUS"]:
         for minsize in range(
             config["baseSizeMUS"], max(config["baseSizeMUS"] + 1, 10000), 1
         ):
+            logging.info("Looking for %s (know %s)", minsize, MUSSizeFound.value)
+            MUSSizeRequired.value = minsize
+            if CONFIG["earlyExit"] and MUSSizeFound.value <= minsize:
+                logging.info("Early exit because MUS already known")
+                return
+
             with getPool(CONFIG["cores"]) as pool:
                 # Do 'range(repeats)' first, so when we distribute we get an
                 # even spread of literals on different cores minsize+1 for MUS
@@ -513,6 +548,7 @@ def cascadeMUS(solver, puzlits, repeats, musdict, config):
             for minsize in range(
                 config["baseSizeMUS"], max(config["baseSizeMUS"] + 1, 10000), 1
             ):
+                MUSSizeRequired.value = minsize
                 # Do 'range(repeats)' first, so when we distribute we get an 
                 # even spread of literals on different cores minsize+1 for MUS 
                 # size, as the MUS will include 'p'.
@@ -568,13 +604,36 @@ class CascadeMUSFinder:
                 musdict,
                 repeats=CONFIG["smallRepeats"],
                 distance=1,
+                badlimit=3
             )
 
+        logging.info("Smallest MUS A: %s ", musdict.minimum())
+
         # Early exit for trivial case
-        if musdict.minimum() == 1:
+        if musdict.minimum() <= 1:
             logging.info("Early exit from checkSmall1")
             return musdict
 
+        # Try looking for general tiny MUSes, to prime search
+        logging.info("Looking for small")
+        getTinyMUSes(
+                self._solver,
+                puzlits,
+                musdict,
+                repeats=CONFIG["smallRepeats"],
+                distance=999,
+                badlimit=CONFIG["baseSizeMUS"]*2
+        )
+
+        logging.info("Smallest MUS B: %s ", musdict.minimum())
+
+        # Early exit for trivial case
+        if musdict.minimum() <= 1:
+            logging.info("Early exit from checkSmall general")
+            return musdict
+
+        logging.info("Checking cache")
+
         if CONFIG["useCache"]:
             checkMUS(self._solver, puzlits, self._bestcache, musdict)
 
@@ -586,9 +645,11 @@ class CascadeMUSFinder:
                 musdict,
                 repeats=CONFIG["smallRepeats"],
                 distance=2,
+                badlimit=5
             )
 
         if not CONFIG["checkSmall2"]:
+            logging.info("Running cascade algorithm")        
             cascadeMUS(
                 self._solver, puzlits, CONFIG["repeats"], musdict, CONFIG
             )
